"""
This module provides tools to read information about upstream mirror, e.g.,
if it has a specific julia download.
"""
from .defaults import default_scheme_ports
from .defaults import SOURCE_CONFIGFILE
from .net_utils import query_ip
from .net_utils import port_response_time
from .net_utils import first_response
from .sys_utils import show_verbose
from .filters import generate_info
from .interactive_utils import color

from itertools import chain, repeat
from urllib.parse import urlparse
from string import Template

import json
import os

from typing import List


class ReleaseSource:
    def __init__(self,
                 name: str,
                 urls: List[str],
                 latest_urls: List[str],
                 timeout=2.0):
        # seperate stable and nightly versions because:
        #   * JuliaComputing stores them in two different s3 buckets
        #   * not all mirror servers need/want to serve nightly releases
        # store multiple urls because:
        #   * many mirrors have different multiple domains for different
        #     network choices, e.g., ipv4, ipv6, cernet, chinanet, rsync
        self.name = name
        self.url_templates = [Template(x) for x in urls]
        # TODO: make latest an optional config
        self.latest_url_templates = [Template(x) for x in latest_urls if x]
        self.timeout = timeout
        self._latencies = dict()  # type: ignore

    @property
    def urls(self):
        return [x.template for x in self.url_templates]

    @property
    def latest_urls(self):
        return [x.template for x in self.latest_url_templates]

    @property
    def hosts(self):
        return [urlparse(url).netloc for url in self.urls]

    @property
    def latest_hosts(self):
        return [urlparse(url).netloc for url in self.latest_urls]

    @property
    def latencies(self):
        # only check latency once and lazily
        if not self._latencies:
            url_list = self.urls.copy()
            url_list.extend(self.latest_urls)
            host_list = [urlparse(url).netloc for url in url_list]

            # hosts might point to the same ip address
            host_ip_records = {host: query_ip(host) for host in host_list}
            ip_port_records = {host_ip_records[urlparse(url).netloc]:
                               default_scheme_ports[urlparse(url).scheme]
                               for url in url_list}

            latency_records = dict()
            # TODO: use threads
            for ip, port in ip_port_records.items():
                latency_records[ip] = port_response_time(ip, port,
                                                         self.timeout)
            self._latencies = {host: latency_records[host_ip_records[host]]
                               for host in host_list}
        return self._latencies

    def __repr__(self):
        return f"ReleaseSource('{self.name}')"

    def get_url(self, plain_version, system, architecture):
        """
        return one potential downloading url with minal network latency for
        specific version, system and architecture. Special version name such
        as 'latest' are treated differently.
        """
        if plain_version == "latest":
            template_lists = self.latest_url_templates
        else:
            template_lists = self.url_templates
        configs = generate_info(plain_version, system, architecture)
        url_list = [t.substitute(**configs) for t in template_lists]
        url_list.sort(key=lambda url:
                      self.latencies[urlparse(url).netloc])
        return url_list[0] if url_list else ""


def read_registry():
    registry = dict()
    for cfg_file in reversed(SOURCE_CONFIGFILE):
        if not os.path.isfile(cfg_file):
            continue
        with open(cfg_file, 'r') as f:
            upstream_records = json.load(f).get("upstream", {})
            temp_registry = {k: ReleaseSource(**v)
                             for k, v in upstream_records.items()}
        registry.update(temp_registry)
    return registry


def latency_string(latency):
    latency *= 1000
    if latency <= 100:
        latency = f"{color.GREEN}{latency:.0f}{color.END}"
    elif latency <= 200:
        latency = f"{color.YELLOW}{latency:.0f}{color.END}"
    else:
        latency = f"{color.RED}{latency:.0f}{color.END}"
    return latency


class SourceRegistry:
    # share the same "full" registry across all instances
    class __SourceRegistry:
        def __init__(self, timeout):
            self.timeout = timeout
            self.registry = read_registry()

    __inner_registry = None

    def __init__(self, *, upstream=None, timeout=2):
        self.upstream = upstream
        if not self.__inner_registry:
            self.__inner_registry = self.__SourceRegistry(timeout)

    @property
    def registry(self):
        if self.upstream:
            # users limit themselves to only one download source
            if self.upstream in self.__inner_registry.registry:
                return {self.upstream:
                        self.__inner_registry.registry[self.upstream]}
            else:
                msg = "valid sources are:" + \
                    ', '.join(self.__inner_registry.registry.keys())
                raise ValueError(msg)
        return self.__inner_registry.registry

    @property
    def latencies(self):
        records = dict()
        for src in self.registry.values():
            records.update(src.latencies)
        return records

    def __len__(self):
        return len(self.registry)

    def info(self):
        msg = f"Found {len(self)} release sources:\n\n"
        for name, resource in self.registry.items():
            msg += f"- {color.BOLD}{name}{color.END}: {resource.name}\n"

            hosts = set(resource.hosts)
            latest_hosts = set(resource.latest_hosts)
            diff = latest_hosts - hosts
            for host in resource.hosts + list(diff):
                latency = latency_string(resource.latencies[host])
                msg += f"  * {host} ({latency} ms)\n"
        return msg

    def _get_urls(self, plain_version, system, architecture):
        """
        return a list of potential downloading urls for specific version,
        system and architecture. Special version name such as 'latest' are
        treated differently.
        """
        url_list = [src.get_url(plain_version, system, architecture)
                    for src in self.registry.values()]
        url_list = [url for url in url_list if url]
        url_list.sort(key=lambda url:
                      self.latencies[urlparse(url).netloc])
        return url_list

    def query_download_url(self,
                           version, system, arch, *,
                           timeout=3):
        """
        return a valid download url to nearest mirror server. If there isn't
        such version then return None.
        """
        url_list = self._get_urls(version, system, arch)
        return first_response(url_list, timeout=timeout)


def show_upstream():
    """print all registered upstream servers"""
    registry = SourceRegistry()
    print(registry.info())


def verify_upstream(registry_name):
    registry = SourceRegistry()
    if registry_name not in registry.registry:
        available_names = '"' + '", "'.join(registry.registry.keys()) + '"'
        raise(ValueError(
            f'registry "{registry_name}" not in {available_names}'))



from .utils.defaults import default_path_template
from .utils.defaults import default_filename_template
from .utils.defaults import default_latest_filename_template
from .utils import generate_info, is_valid_release
from .utils import update_releases
from .utils import read_releases
from .utils import current_system
from .download import download_package

from string import Template

import semantic_version

import json
import os
import logging
import time


class MirrorConfig:
    def __init__(self, configfile, outdir):
        if current_system() == "windows":
            # Windows users (e.g., me) sometimes confuse the use of \\ and \
            outdir = outdir.replace("\\\\", "\\")
        self.configfile = os.path.abspath(os.path.expanduser(configfile))
        self.outdir = outdir

    @property
    def config(self):
        if not os.path.isfile(self.configfile):
            return {}
        with open(self.configfile, 'r') as f:
            return json.load(f)

    @property
    def require_latest(self):
        """True to mirror nightly build as well"""
        return self.config.get("require_latest", True)

    @property
    def overwrite(self):
        """
        True to overwrite existing stable releases as well.
        Nightly builds will be overwrite always regardless of this setting.
        """
        return self.config.get("overwrite", False)

    @property
    def path_template(self):
        """path template to define the folder structure of releases"""
        return Template(self.config.get("path", default_path_template))

    @property
    def filename_template(self):
        """filename template for stable releases"""
        return Template(self.config.get("filename", default_filename_template))

    @property
    def latest_filename_template(self):
        """filename template for nightly builds"""
        return Template(self.config.get("latest_filename",
                                        default_latest_filename_template))

    @property
    def version(self):
        versions = list(set(map(lambda x: x[0],
                                read_releases(stable_only=True))))
        # not using our extended Version
        versions.sort(key=lambda ver: semantic_version.Version(ver))
        if self.require_latest:
            versions.append("latest")
        return versions

    @property
    def system(self):
        return set(map(lambda x: x[1], read_releases()))

    @property
    def architecture(self):
        return set(map(lambda x: x[2], read_releases()))

    @property
    def releases(self):
        stable_only = False if self.require_latest else True
        return read_releases(stable_only)

    def logging(self):
        logging.info(f"mirror configuration:")
        logging.info(f"    - configfile: {self.configfile}")
        logging.info(f"    - outdir: {self.outdir}")
        logging.info(f"    - path: {self.path_template.template}")
        logging.info(f"    - filename: {self.filename_template.template}")
        logging.info(f"    - versions: {', '.join(self.version)}")
        logging.info(f"    - systems: {', '.join(self.system)}")
        logging.info(f"    - architectures: {', '.join(self.architecture)}")
        logging.info(f"    - overwrite: {self.overwrite}")
        logging.info(f"    - require_latest: {self.require_latest}")

    def get_outpath(self, plain_version, system, architecture):
        configs = generate_info(plain_version, system, architecture)
        if plain_version in ["latest"]:
            t_file = self.latest_filename_template
        else:
            t_file = self.filename_template
        configs["filename"] = t_file.substitute(**configs)
        return self.path_template.substitute(**configs)


class Mirror:
    def __init__(self,  config):
        if not isinstance(config, MirrorConfig):
            config = MirrorConfig(config)
        self.config = config

    def pull_releases(self, *,
                      upstream=None):
        for item in self.config.releases:
            filepath = self.config.get_outpath(*item)
            outpath = os.path.join(self.config.outdir, filepath)
            outdir, filename = os.path.split(outpath)

            logging.info(f"start to pull {filepath}")
            overwrite = True if item[0] == "latest" else self.config.overwrite
            download_package(*item,
                             outdir=outdir,
                             upstream=upstream,
                             overwrite=overwrite)


def mirror(outdir="julia_pkg", *,
           period=0,
           upstream=None,
           logfile="mirror.log",
           config="mirror.json"):
    """
    Download/sync all Julia releases

    1. checks if there're new julia releases
    2. downloads all releases Julia releases into `outdir` (default `./julia_pkg`)
    3. (Optional): with flag `--period PERIOD`, it will repeat step 1 and 2 every `PERIOD` seconds

    If you want to modify the default mirror configuration, then provide
    a `mirror.json` file and pass the path to `config`. By default it's at
    the current directory. A template can be found at [1]

    Repository [jill-mirror](https://github.com/johnnychen94/julia-mirror) provides an easy to
    start `docker-compose.yml` for you to start with, which is just a simple docker image built
    upon `jill mirror`.

    [1]: https://github.com/johnnychen94/jill.py/blob/master/mirror.example.json # nopep8

    Arguments:
      outdir: default 'julia_pkg'.
      period: the time between two sync operation. 0(default) to sync once.
      upstream:
        manually choose a download upstream. For example, set it to "Official"
        if you want to download from JuliaComputing's s3 buckets.
      config:
        path to mirror config file
      logfile:
        path to mirror log file
    """
    log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    period = int(period)
    upstream = None if upstream == "None" else upstream

    logger = logging.getLogger('')
    fh = logging.FileHandler(logfile)
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter(log_format)
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    # TODO: filter out urllib3 debug logs

    m = Mirror(MirrorConfig(config, outdir=outdir))
    m.config.logging()
    while True:
        update_releases(system="all", architecture="all")

        logging.info("START: pull Julia releases")
        m.pull_releases(upstream=upstream)
        logging.info("END: pulling Julia releases")

        if period == 0:
            return True
        else:
            time.sleep(period)

        # refresh configuration at each re-pull
        logging.info("reload configure file")
        m.config = MirrorConfig(config, outdir=outdir)
        m.config.logging()



from .utils import SourceRegistry, verify_upstream
from .utils import latest_version
from .utils import is_version_released
from .utils import is_full_version
from .utils import current_system, current_architecture
from .utils import verify_gpg
from .utils import color

import wget
import os
import shutil
import tempfile
import logging

from urllib.parse import urlparse

from typing import Optional

from urllib.error import URLError


def _download(url: str, out: str):
    # always do overwrite
    outpath = os.path.abspath(out)
    outdir, outname = os.path.split(outpath)

    with tempfile.TemporaryDirectory() as temp_outdir:
        temp_outpath = os.path.join(temp_outdir, outname)
        try:
            msg = f"downloading from {url}"
            logging.info(msg)
            print(msg)
            wget.download(url, temp_outpath)
            print()  # for format usage
            msg = f"finished downloading {outname}"
            print(f"{color.GREEN}{msg}{color.END}")
        except (URLError, ConnectionError) as e:
            msg = f"failed to download {outname}"
            logging.info(msg)
            print(f"{color.RED}{msg}{color.END}")
            return False

        if not os.path.isdir(outdir):
            os.makedirs(outdir, exist_ok=True)
        shutil.move(temp_outpath, outpath)

    return outpath


def download_package(version=None, sys=None, arch=None, *,
                     upstream=None,
                     outdir=None,
                     overwrite=False):
    """
    download julia release from nearest servers

    `jill download [version] [sys] [arch]` downloads a Julia release
    in your current folder. If you don't specify any argument, then
    it will download the latest stable version for your current platform.

    The syntax for `version` is:

    * `stable`: latest stable Julia release. This is the _default_ option.
    * `1`: latest `1.y.z` Julia release.
    * `1.0`: latest `1.0.z` Julia release.
    * `1.4.0-rc1`: as it is. This is the only way to install unstable release.
    * `latest`/`nightly`: the nightly builds from source code.

    For whatever reason, if you only want to download release from
    a specific upstream (e.g., from JuliaComputing), then you can use
    `--upstream` flag (e.g., `jill download --upstream Official`).

    To see a full list of upstream servers, please use `jill upstream`.

    If you're interested in downloading from an unregistered private
    mirror, you can provide a `sources.json` file to CONFIG_PATH and use
    `jill upstream` to check if your mirror is added. A config template
    can be found at [1]. For how to make such mirror, please refer to
    `jill mirror`.

    CONFIG_PATH:
      * windows: `~\\AppData\\Local\\julias\\sources.json`
      * other: `~/.config/jill/sources.json`

    [1]: https://github.com/johnnychen94/jill.py/blob/master/jill/config/sources.json # nopep8

    Arguments:
      version:
        The Julia version you want to install. See also `jill install`
      sys: Options are: "linux", "macos", "freebsd", "windows"
      arch: Options are: "i686", "x86_64", "ARMv7", "ARMv8"
      upstream:
        manually choose a download upstream. For example, set it to "Official"
        if you want to download from JuliaComputing's s3 buckets.
      outdir:
        where release is downloaded to. By default it's the current folder.
      overwrite:
        add `--overwrite` flag to overwrite existing releases.
    """
    version = str(version) if (version or str(version) == "0") else ""
    version = "latest" if version == "nightly" else version
    version = "" if version == "stable" else version

    system = sys if sys else current_system()
    architecture = arch if arch else current_architecture()

    # allow downloading unregistered releases, e.g., 1.4.0-rc1
    do_release_check = not is_full_version(version)

    if upstream:
        verify_upstream(upstream)
    wrong_args = False
    try:
        version = latest_version(
            version, system, architecture, update=True, upstream=upstream)
    except ValueError as e:
        # hide the nested error stack :P
        wrong_args = True
    if wrong_args:
        msg = f"something wrong for the platform argument you passed:\n"
        msg += f"  - version(>= 0.6.0): {version}\n"
        msg += f"  - system: {system}\n"
        msg += f"  - archtecture: {architecture}\n"
        msg += f"example: `jill download 1 linux x86_64`"
        raise(ValueError(msg))

    release_str = f"{version}-{system}-{architecture}"
    if do_release_check:
        rst = is_version_released(
            version, system, architecture, upstream=upstream)
        if not rst:
            msg = f"failed to find Julia release for {release_str}."
            logging.info(msg)
            print(f"{color.RED}{msg}{color.END}")
            return False

    msg = f"downloading Julia release for {release_str}"
    logging.info(msg)
    print(msg)
    registry = SourceRegistry(upstream=upstream)
    url = registry.query_download_url(version, system, architecture)
    if not url:
        msg = f"failed to find available upstream for {release_str}"
        logging.warning(msg)
        print(f"{color.RED}{msg}{color.END}")
        return None

    outdir = outdir if outdir else '.'
    outdir = os.path.abspath(outdir)
    if current_system() == "windows":
        outdir = outdir.replace("\\\\", "\\")

    outname = os.path.split(urlparse(url).path)[1]
    outpath = os.path.join(outdir, outname)

    skip_download = False
    if system in ["linux", "freebsd"]:
        if (os.path.isfile(outpath) and os.path.isfile(outpath+".asc")
                and not overwrite):
            skip_download = True
    else:
        if os.path.isfile(outpath) and not overwrite:
            skip_download = True
    if skip_download:
        msg = f"{outname} already exists, skip downloading"
        logging.info(msg)
        print(f"{color.GREEN}{msg}{color.END}")
        return outpath

    package_path = _download(url, outpath)

    if system in ["windows", "macos"]:
        # macOS and Windows releases are codesigned with certificates
        # that are verified by the operating system during installation
        return package_path
    elif system in ["linux", "freebsd"]:
        # need additional verification using GPG
        if not package_path:
            return package_path

        # a mirror should provides both *.tar.gz and *.tar.gz.asc
        gpg_signature_path = _download(url+".asc", outpath+".asc")
        if not gpg_signature_path:
            msg = f"failed to download GPG signature for {release_str}\n"
            msg += "remove untrusted/broken file"
            logging.info(msg)
            print(f"{color.RED}{msg}{color.END}")
            os.remove(package_path)
            return False

        if not verify_gpg(package_path, gpg_signature_path):
            msg = f"failed to verify {release_str} downloads\n"
            msg += "remove untrusted/broken files"
            logging.info(msg)
            print(f"{color.RED}{msg}{color.END}")
            os.remove(package_path)
            os.remove(gpg_signature_path)
            return False

        # GPG-verified julia release path
        msg = f"verifying {release_str} succeed"
        logging.info(msg)
        print(f"{color.GREEN}{msg}{color.END}")
        return package_path
    else:
        raise ValueError(f"unsupported system {sys}")
